/**
  Title: Raspberry Pi Pico Touch Panel
  Author: Makoto Ito
  Date: 30-11-2023

  Using Earle Philhower's 'arduino-pico': https://github.com/earlephilhower/arduino-pico . This core comes with LittleFS library.
  Make sure to flash your Pico with at least 64KB for the filesystem!

  This example is based on the examples that come with the TFT_eSPI library. Which includes the Adafruit GFX button handling class.

  These libraries need to be installed:

  - TFT_eSPI (available through the Library Manager)

  ! Don't forget to configure User_Setup.h in the TFT_eSPI library!
  
    TFT_MISO  0
    TFT_MOSI  3
    TFT_SCLK  2
    TFT_CS   20
    TFT_DC   22
    TFT_RST  21
    TOUCH_CS 14

*/

#include <FS.h>
#include <LittleFS.h>
#include <SPI.h>
#include <TFT_eSPI.h>

#define CALIBRATION_FILE "caribration.txt"            // Touch calibration file
#define REPEAT_CAL false                              // Set to tru if you want the calibration to run on each boot

#define KEY_W 95                                      // Key width
#define KEY_H 95                                      // Key height
#define KEY_SPACING_X 20 // X gap
#define KEY_SPACING_Y 10 // Y gap
#define KEY_X (KEY_W/2) + KEY_SPACING_X               // X-axis centre of the first key
#define KEY_Y (KEY_H/2) + KEY_SPACING_Y               // Y-axis centre of the first key
#define KEY_TEXTSIZE 3                                // Font size multiplier
#define LABEL1_FONT &FreeSansOblique12pt7b            // Key label font

uint8_t keydelay = 100;
TFT_eSPI tft = TFT_eSPI();                            // Create the screen object

char keyLabel[12][3] = {"1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"};
uint16_t keyColor[12] = {TFT_BLUE, TFT_RED, TFT_GREEN, TFT_NAVY, 
                         TFT_MAROON, TFT_MAGENTA, TFT_ORANGE, TFT_SKYBLUE, 
                         TFT_PURPLE, TFT_CYAN, TFT_PINK, TFT_DARKCYAN
                        };
TFT_eSPI_Button key[12];

int dispPage = 0;
int lastButton = 0;
int version = 1;

/*!
 @fn
 @brief システム初期化（起動時にシステムから一度だけ呼ばれる）
 @param   なし
 @return  なし
 */
void setup()
{
  Serial.begin(115200);
  Serial.println("");

  Serial2.begin(9600, SERIAL_8N1);
  Serial2.setTimeout(20);
  
  Serial2.println("");
  Serial2.println("");
  Serial2.println("Pico Systems LCD Touch Panel.");
  Serial2.println("");

  tft.init();
  tft.setRotation(1);
  touch_calibrate(false);
  dispPage = 1;
  drawScreen(dispPage);
  lastButton = 0;
}

/*!
 @fn
 @brief メインループ（システムから定期的に呼ばれる）
 @param   なし
 @return  なし
 */
void loop()
{
  uint16_t t_x = 0, t_y = 0;
  bool pressed;
  char buf[256];
  int len;
  char tmp, cmnd;
  int param;
  char str[16];
  static int i;
  char num[4];

  if(dispPage == 1)                                   // support only page 1 now
  {
    pressed = tft.getTouch(&t_x, &t_y);               // check each key pressed
    for (uint8_t b = 0; b < 12; b++)
    {
      if (pressed && key[b].contains(t_x, t_y))
        key[b].press(true);                           //pressed
      else
        key[b].press(false);                          // NOT pressed
    }

    for (uint8_t b = 0; b < 12; b++)                  // Check if any key has changed state
    {
      if (key[b].justReleased()) key[b].drawButton(); // draw normal again

      if (key[b].justPressed()) 
      {
        key[b].drawButton(true);                      // draw invert
        buttonpress(b);                               // Call the button press function
      }
    }
  }

  tmp = Serial2.read();
  if(tmp != 0xff)
  {
    buf[i] = tmp;
    if(i < 255) i++;
  
    if(tmp == '\r')                                   // <CR>
    {
      if(i >= 3)                                      // valid frame
      {
        cmnd = '\0';
        param = 0;
        if(isupper(buf[i - 4]))
        {
          cmnd = buf[i - 4];
          if((isdigit(buf[i - 3])) && (isdigit(buf[i - 2])))
            param = (buf[i - 3] - 0x30) * 10 + (buf[i - 2] - 0x30);
          else                                        // bad parameter
            cmnd = '\0';

          switch(cmnd)
          {
            case 'D':
                dispPage = param;
                drawScreen(dispPage);
                lastButton = 0;
              break;

            case 'R':
                rp2040.reboot();                      // force reset
              break;

            case 'I':
                if(param == 0)                        // get version
                {
                  str[0] = 'V';                       // V event as response
                  sprintf(&str[1], "%02d", version);
                  str[3] = '\r';
                  str[4] = '\0';
                  Serial2.print(str);
                }
                else if(param == 1)                   // get page number
                {
                  str[0] = 'P';                       // P event as response
                  sprintf(&str[1], "%02d", dispPage);
                  str[3] = '\r';
                  str[4] = '\0';
                  Serial2.print(str);
                }
                else if(param == 2)                   // get last button
                {
                  str[0] = 'L';                       // L event as response
                  sprintf(&str[1], "%02d", lastButton);
                  str[3] = '\r';
                  str[4] = '\0';
                  Serial2.print(str);
                }
              break;

            case 'C':
                touch_calibrate(true);
                drawScreen(dispPage);

                strcpy(str, "F00\r");                 // F Event as response
                Serial2.print(str);
              break;

            default:
              break;
          }
        }
      }
      i = 0;
    }
  }
}

/*!
 @fn
 @brief 画面描画
 @param   page  表示ページ  0で非表示
 @return  なし
 */
void drawScreen(int page)
{
  tft.fillScreen(TFT_BLACK);

  if(page == 1)                                       // support only page 1 now
  {
    // Draw the keys ( 3 times 4 loops to create 12)
    for (uint8_t row = 0; row < 3; row++)
    {
      for (uint8_t col = 0; col < 4; col++)
      {
        uint8_t b = col + row * 4;                    // The button number. The first button is 0.
        key[b].initButton(&tft,                       // x, y, w, h, outlinecolor, fillcolor, textcolor, keyLabel, textsize
                          KEY_X + col * (KEY_W + KEY_SPACING_X),
                          KEY_Y + row * (KEY_H + KEY_SPACING_Y),
                          KEY_W,
                          KEY_H,
                          TFT_WHITE,
                          keyColor[b],
                          TFT_WHITE,
                          keyLabel[b],
                          KEY_TEXTSIZE);
        key[b].drawButton();
      }
    }
  }
}

/*!
 @fn
 @brief ボタンが押された時の処理
 @param [in]  button  ボタン番号（0 - 99）
 @return  なし
 */
void buttonpress(int button)
{
  char str[32];
/*                                                  // Debug print
  strcpy(str, "Button ");
  sprintf(num, "%d", button + 1);
  strcat(str, num);
  strcat(str, " pressed");
  Serial.println(str);
*/
  str[0] = 'B';
  sprintf(&str[1], "%02d", button);
  str[3] = '\r';
  str[4] = '\0';
  Serial2.print(str);                                 // COM out
  lastButton = button;
}

/*!
 @fn
 @brief タッチパネル校正処理
 @param   force trueで強制的に校正、falseで通常動作
 @return  なし
 */
void touch_calibrate(bool force)
{
  File datafile;
  uint16_t calData[5];
  uint8_t calDataOK = 0;
  int len;

  LittleFS.begin();

  if(force) LittleFS.remove(CALIBRATION_FILE);

  if(LittleFS.exists(CALIBRATION_FILE))
  {
    if(REPEAT_CAL)
      LittleFS.remove(CALIBRATION_FILE);
    else
    {                                                 // read calibration data
      datafile = LittleFS.open(CALIBRATION_FILE, "r");
      if(datafile)
      {
        len = datafile.readBytes((char*)calData, 10);
        if(len == 10) calDataOK = 1;
        datafile.close();
      }
    }
  }

  if (calDataOK && !REPEAT_CAL)                       // calibration data valid
    tft.setTouch(calData);
  else
  {                                                   // data not valid so recalibrate
    tft.fillScreen(TFT_BLACK);
    tft.setCursor(20, 0);
    tft.setTextFont(2);
    tft.setTextSize(1);
    tft.setTextColor(TFT_WHITE, TFT_BLACK);

    tft.println("Touch corners as indicated");

    tft.setTextFont(1);
    tft.println();

    if(REPEAT_CAL)
    {
      tft.setTextColor(TFT_RED, TFT_BLACK);
      tft.println("Set REPEAT_CAL to false to stop this running again!");
    }

    tft.calibrateTouch(calData, TFT_MAGENTA, TFT_BLACK, 15);

    tft.setTextColor(TFT_GREEN, TFT_BLACK);
    tft.println("Calibration complete!");

    datafile = LittleFS.open(CALIBRATION_FILE, "w");
    if(datafile)                                      // store calibration data
    {
      datafile.write((const unsigned char *)calData, 10);
      datafile.close();
    }
  }

  LittleFS.end();
}
